@@ -4,6 +4,10 @@
 //
 // Copyright (c) 2011-2017 Adam Wulkiewicz, Lodz, Poland.
 //
+// This file was modified by Oracle on 2019-2020.
+// Modifications copyright (c) 2019-2020 Oracle and/or its affiliates.
+// Contributed and/or modified by Adam Wulkiewicz, on behalf of Oracle
+//
 // Use, modification and distribution is subject to the Boost Software License,
 // Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
 // http://www.boost.org/LICENSE_1_0.txt)
@@ -13,6 +17,8 @@
 
 #include <boost/core/ignore_unused.hpp>
 
+#include <boost/geometry/core/static_assert.hpp>
+
 #include <boost/geometry/index/detail/algorithms/intersection_content.hpp>
 #include <boost/geometry/index/detail/algorithms/margin.hpp>
 #include <boost/geometry/index/detail/algorithms/nth_element.hpp>
@@ -30,23 +36,27 @@ namespace detail { namespace rtree {
 
 namespace rstar {
 
-template <typename Element, typename Translator, typename Tag, size_t Corner, size_t AxisIndex>
+template <typename Element, typename Parameters, typename Translator, typename Tag, size_t Corner, size_t AxisIndex>
 class element_axis_corner_less
 {
     typedef typename rtree::element_indexable_type<Element, Translator>::type indexable_type;
     typedef typename geometry::point_type<indexable_type>::type point_type;
     typedef geometry::model::box<point_type> bounds_type;
-    typedef index::detail::bounded_view<indexable_type, bounds_type> bounded_view_type;
+    typedef typename index::detail::strategy_type<Parameters>::type strategy_type;
+    typedef index::detail::bounded_view
+        <
+            indexable_type, bounds_type, strategy_type
+        > bounded_view_type;
 
 public:
-    element_axis_corner_less(Translator const& tr)
-        : m_tr(tr)
+    element_axis_corner_less(Translator const& tr, strategy_type const& strategy)
+        : m_tr(tr), m_strategy(strategy)
     {}
 
     bool operator()(Element const& e1, Element const& e2) const
     {
-        bounded_view_type bounded_ind1(rtree::element_indexable(e1, m_tr));
-        bounded_view_type bounded_ind2(rtree::element_indexable(e2, m_tr));
+        bounded_view_type bounded_ind1(rtree::element_indexable(e1, m_tr), m_strategy);
+        bounded_view_type bounded_ind2(rtree::element_indexable(e2, m_tr), m_strategy);
 
         return geometry::get<Corner, AxisIndex>(bounded_ind1)
             < geometry::get<Corner, AxisIndex>(bounded_ind2);
@@ -54,13 +64,16 @@ public:
 
 private:
     Translator const& m_tr;
+    strategy_type const& m_strategy;
 };
 
-template <typename Element, typename Translator, size_t Corner, size_t AxisIndex>
-class element_axis_corner_less<Element, Translator, box_tag, Corner, AxisIndex>
+template <typename Element, typename Parameters, typename Translator, size_t Corner, size_t AxisIndex>
+class element_axis_corner_less<Element, Parameters, Translator, box_tag, Corner, AxisIndex>
 {
+    typedef typename index::detail::strategy_type<Parameters>::type strategy_type;
+
 public:
-    element_axis_corner_less(Translator const& tr)
+    element_axis_corner_less(Translator const& tr, strategy_type const&)
         : m_tr(tr)
     {}
 
@@ -74,11 +87,13 @@ private:
     Translator const& m_tr;
 };
 
-template <typename Element, typename Translator, size_t Corner, size_t AxisIndex>
-class element_axis_corner_less<Element, Translator, point_tag, Corner, AxisIndex>
+template <typename Element, typename Parameters, typename Translator, size_t Corner, size_t AxisIndex>
+class element_axis_corner_less<Element, Parameters, Translator, point_tag, Corner, AxisIndex>
 {
+    typedef typename index::detail::strategy_type<Parameters>::type strategy_type;
+
 public:
-    element_axis_corner_less(Translator const& tr)
+    element_axis_corner_less(Translator const& tr, strategy_type const& )
         : m_tr(tr)
     {}
 
@@ -113,6 +128,9 @@ struct choose_split_axis_and_index_for_corner
 
         BOOST_GEOMETRY_INDEX_ASSERT(elements.size() == parameters.get_max_elements() + 1, "wrong number of elements");
 
+        typename index::detail::strategy_type<Parameters>::type const&
+            strategy = index::detail::get_strategy(parameters);
+
         // copy elements
         Elements elements_copy(elements);                                                                       // MAY THROW, STRONG (alloc, copy)
         
@@ -120,7 +138,10 @@ struct choose_split_axis_and_index_for_corner
         size_t const index_last = parameters.get_max_elements() - parameters.get_min_elements() + 2;
 
         // sort elements
-        element_axis_corner_less<element_type, Translator, indexable_tag, Corner, AxisIndex> elements_less(translator);
+        element_axis_corner_less
+            <
+                element_type, Parameters, Translator, indexable_tag, Corner, AxisIndex
+            > elements_less(translator, strategy);
         std::sort(elements_copy.begin(), elements_copy.end(), elements_less);                                   // MAY THROW, BASIC (copy)
 //        {
 //            typename Elements::iterator f = elements_copy.begin() + index_first;
@@ -143,12 +164,14 @@ struct choose_split_axis_and_index_for_corner
             // TODO - awulkiew: may be optimized - box of group 1 may be initialized with
             // box of min_elems number of elements and expanded for each iteration by another element
 
-            Box box1 = rtree::elements_box<Box>(elements_copy.begin(), elements_copy.begin() + i, translator);
-            Box box2 = rtree::elements_box<Box>(elements_copy.begin() + i, elements_copy.end(), translator);
+            Box box1 = rtree::elements_box<Box>(elements_copy.begin(), elements_copy.begin() + i,
+                                                translator, strategy);
+            Box box2 = rtree::elements_box<Box>(elements_copy.begin() + i, elements_copy.end(),
+                                                translator, strategy);
             
             sum_of_margins += index::detail::comparable_margin(box1) + index::detail::comparable_margin(box2);
 
-            content_type ovl = index::detail::intersection_content(box1, box2);
+            content_type ovl = index::detail::intersection_content(box1, box2, strategy);
             content_type con = index::detail::content(box1) + index::detail::content(box2);
 
             // TODO - shouldn't here be < instead of <= ?
@@ -167,7 +190,7 @@ struct choose_split_axis_and_index_for_corner
 //template <typename Box, size_t AxisIndex, typename ElementIndexableTag>
 //struct choose_split_axis_and_index_for_axis
 //{
-//    BOOST_MPL_ASSERT_MSG(false, NOT_IMPLEMENTED_FOR_THIS_TAG, (ElementIndexableTag));
+//    BOOST_GEOMETRY_STATIC_ASSERT_FALSE("Not implemented for this Tag type.", ElementIndexableTag);
 //};
 
 template <typename Box, size_t AxisIndex, typename ElementIndexableTag>
@@ -337,14 +360,15 @@ struct nth_element
     BOOST_STATIC_ASSERT(0 < Dimension);
     BOOST_STATIC_ASSERT(I < Dimension);
 
-    template <typename Elements, typename Translator>
-    static inline void apply(Elements & elements, const size_t axis, const size_t index, Translator const& tr)
+    template <typename Elements, typename Parameters, typename Translator>
+    static inline void apply(Elements & elements, Parameters const& parameters,
+                             const size_t axis, const size_t index, Translator const& tr)
     {
         //BOOST_GEOMETRY_INDEX_ASSERT(axis < Dimension, "unexpected axis value");
 
         if ( axis != I )
         {
-            nth_element<Corner, Dimension, I + 1>::apply(elements, axis, index, tr);                          // MAY THROW, BASIC (copy)
+            nth_element<Corner, Dimension, I + 1>::apply(elements, parameters, axis, index, tr);                     // MAY THROW, BASIC (copy)
         }
         else
         {
@@ -352,7 +376,13 @@ struct nth_element
             typedef typename rtree::element_indexable_type<element_type, Translator>::type indexable_type;
             typedef typename tag<indexable_type>::type indexable_tag;
 
-            element_axis_corner_less<element_type, Translator, indexable_tag, Corner, I> less(tr);
+            typename index::detail::strategy_type<Parameters>::type
+                strategy = index::detail::get_strategy(parameters);
+
+            element_axis_corner_less
+                <
+                    element_type, Parameters, Translator, indexable_tag, Corner, I
+                > less(tr, strategy);
             index::detail::nth_element(elements.begin(), elements.begin() + index, elements.end(), less);            // MAY THROW, BASIC (copy)
         }
     }
@@ -361,36 +391,40 @@ struct nth_element
 template <size_t Corner, size_t Dimension>
 struct nth_element<Corner, Dimension, Dimension>
 {
-    template <typename Elements, typename Translator>
-    static inline void apply(Elements & /*elements*/, const size_t /*axis*/, const size_t /*index*/, Translator const& /*tr*/)
+    template <typename Elements, typename Parameters, typename Translator>
+    static inline void apply(Elements & /*elements*/, Parameters const& /*parameters*/,
+                             const size_t /*axis*/, const size_t /*index*/, Translator const& /*tr*/)
     {}
 };
 
 } // namespace rstar
 
-template <typename Value, typename Options, typename Translator, typename Box, typename Allocators>
-struct redistribute_elements<Value, Options, Translator, Box, Allocators, rstar_tag>
+template <typename MembersHolder>
+struct redistribute_elements<MembersHolder, rstar_tag>
 {
-    typedef typename rtree::node<Value, typename Options::parameters_type, Box, Allocators, typename Options::node_tag>::type node;
-    typedef typename rtree::internal_node<Value, typename Options::parameters_type, Box, Allocators, typename Options::node_tag>::type internal_node;
-    typedef typename rtree::leaf<Value, typename Options::parameters_type, Box, Allocators, typename Options::node_tag>::type leaf;
+    typedef typename MembersHolder::box_type box_type;
+    typedef typename MembersHolder::parameters_type parameters_type;
+    typedef typename MembersHolder::translator_type translator_type;
+    typedef typename MembersHolder::allocators_type allocators_type;
 
-    typedef typename Options::parameters_type parameters_type;
+    typedef typename MembersHolder::node node;
+    typedef typename MembersHolder::internal_node internal_node;
+    typedef typename MembersHolder::leaf leaf;
 
-    static const size_t dimension = geometry::dimension<Box>::value;
+    static const size_t dimension = geometry::dimension<box_type>::value;
 
-    typedef typename index::detail::default_margin_result<Box>::type margin_type;
-    typedef typename index::detail::default_content_result<Box>::type content_type;
+    typedef typename index::detail::default_margin_result<box_type>::type margin_type;
+    typedef typename index::detail::default_content_result<box_type>::type content_type;
 
     template <typename Node>
     static inline void apply(
         Node & n,
         Node & second_node,
-        Box & box1,
-        Box & box2,
+        box_type & box1,
+        box_type & box2,
         parameters_type const& parameters,
-        Translator const& translator,
-        Allocators & allocators)
+        translator_type const& translator,
+        allocators_type & allocators)
     {
         typedef typename rtree::elements_type<Node>::type elements_type;
         typedef typename elements_type::value_type element_type;
@@ -417,7 +451,7 @@ struct redistribute_elements<Value, Options, Translator, Box, Allocators, rstar_
         //       and again, the same below calling partial_sort/nth_element
         //       It would be even possible to not re-sort/find nth_element if the axis/corner
         //       was found for the last sorting - last combination of axis/corner
-        rstar::choose_split_axis_and_index<Box, dimension>
+        rstar::choose_split_axis_and_index<box_type, dimension>
             ::apply(elements_copy,
                     split_axis, split_corner, split_index,
                     smallest_sum_of_margins, smallest_overlap, smallest_content,
@@ -432,23 +466,28 @@ struct redistribute_elements<Value, Options, Translator, Box, Allocators, rstar_
         if ( split_corner == static_cast<size_t>(min_corner) )
         {
             rstar::nth_element<min_corner, dimension>
-                ::apply(elements_copy, split_axis, split_index, translator);                            // MAY THROW, BASIC (copy)
+                ::apply(elements_copy, parameters, split_axis, split_index, translator);                // MAY THROW, BASIC (copy)
         }
         else
         {
             rstar::nth_element<max_corner, dimension>
-                ::apply(elements_copy, split_axis, split_index, translator);                            // MAY THROW, BASIC (copy)
+                ::apply(elements_copy, parameters, split_axis, split_index, translator);                // MAY THROW, BASIC (copy)
         }
 
         BOOST_TRY
         {
+            typename index::detail::strategy_type<parameters_type>::type const&
+                strategy = index::detail::get_strategy(parameters);
+
             // copy elements to nodes
             elements1.assign(elements_copy.begin(), elements_copy.begin() + split_index);               // MAY THROW, BASIC
             elements2.assign(elements_copy.begin() + split_index, elements_copy.end());                 // MAY THROW, BASIC
 
             // calculate boxes
-            box1 = rtree::elements_box<Box>(elements1.begin(), elements1.end(), translator);
-            box2 = rtree::elements_box<Box>(elements2.begin(), elements2.end(), translator);
+            box1 = rtree::elements_box<box_type>(elements1.begin(), elements1.end(),
+                                                 translator, strategy);
+            box2 = rtree::elements_box<box_type>(elements2.begin(), elements2.end(),
+                                                 translator, strategy);
         }
         BOOST_CATCH(...)
         {
@@ -456,7 +495,7 @@ struct redistribute_elements<Value, Options, Translator, Box, Allocators, rstar_
             elements1.clear();
             elements2.clear();
 
-            rtree::destroy_elements<Value, Options, Translator, Box, Allocators>::apply(elements_backup, allocators);
+            rtree::destroy_elements<MembersHolder>::apply(elements_backup, allocators);
             //elements_backup.clear();
 
             BOOST_RETHROW                                                                                 // RETHROW, BASIC
